/*

Copyright (c) 2024 Byron Lathi

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/

// Language: Verilog 2001

`resetall
`timescale 1ns / 1ps
`default_nettype none

/*
 * Ethernet frame duplicator
 */
module eth_duplicator #
(
    parameter DATA_WIDTH = 8,
    parameter KEEP_ENABLE = (DATA_WIDTH>8),
    parameter KEEP_WIDTH = (DATA_WIDTH/8),
    parameter ID_ENABLE = 0,
    parameter ID_WIDTH = 8,
    parameter DEST_ENABLE = 0,
    parameter DEST_WIDTH = 8,
    parameter USER_ENABLE = 1,
    parameter USER_WIDTH = 1
)
(
    input  wire                          clk,
    input  wire                          rst,

    /*
     * Ethernet frame input
     */
    input  wire                          s_eth_hdr_valid,
    output wire                          s_eth_hdr_ready,
    input  wire [47:0]                   s_eth_dest_mac,
    input  wire [47:0]                   s_eth_src_mac,
    input  wire [15:0]                   s_eth_type,
    input  wire [DATA_WIDTH-1:0]         s_eth_payload_axis_tdata,
    input  wire [KEEP_WIDTH-1:0]         s_eth_payload_axis_tkeep,
    input  wire                          s_eth_payload_axis_tvalid,
    output wire                          s_eth_payload_axis_tready,
    input  wire                          s_eth_payload_axis_tlast,
    input  wire [USER_WIDTH-1:0]         s_eth_payload_axis_tuser,
    
    /*
     * Ethernet frame output
     */
    output wire                         m_eth_hdr_valid,
    input  wire                         m_eth_hdr_ready,
    output wire [47:0]                  m_eth_dest_mac,
    output wire [47:0]                  m_eth_src_mac,
    output wire [15:0]                  m_eth_type,
    output wire [DATA_WIDTH-1:0]        m_eth_payload_axis_tdata,
    output wire [KEEP_WIDTH-1:0]        m_eth_payload_axis_tkeep,
    output wire                         m_eth_payload_axis_tvalid,
    input  wire                         m_eth_payload_axis_tready,
    output wire                         m_eth_payload_axis_tlast,
    output wire                         m_eth_payload_axis_tuser,


    /*
     * Frame filter inputs
     */
    input  wire [47:0]                  source_mac_filter,
    input  wire [47:0]                  source_mac_filter_mask,
    input  wire                         source_mac_filter_en,
    input  wire [47:0]                  dest_mac_filter,
    input  wire [47:0]                  dest_mac_filter_mask,
    input  wire                         dest_mac_filter_en,
    input  wire [15:0]                  ethertype_filter,
    input  wire [15:0]                  ethertype_filter_mask,
    input  wire                         ethertype_filter_en,

    /*
     * Frame mod inputs
     */
    input  wire [47:0]                  source_mac_mod,
    input  wire [47:0]                  source_mac_mod_mask,
    input  wire                         source_mac_mod_en,
    input  wire [47:0]                  dest_mac_mod,
    input  wire [47:0]                  dest_mac_mod_mask,
    input  wire                         dest_mac_mod_en,
    input  wire [15:0]                  ethertype_mod,
    input  wire [15:0]                  ethertype_mod_mask,
    input  wire                         ethertype_mod_en
);

reg [47:0]  source_mac_reg, source_mac_next;
reg [47:0]  dest_mac_reg, dest_mac_next;
reg [15:0]  ethertype_reg, ethertype_next;

reg [47:0]  source_mac_modified;
reg [47:0]  dest_mac_modified;
reg [15:0]  ethertype_modified;

reg s_eth_hdr_ready_int;
reg s_eth_payload_axis_tready_int;

assign s_eth_hdr_ready = s_eth_hdr_ready_int;
assign s_eth_payload_axis_tready = s_eth_payload_axis_tready_int;

reg hdr_ready_seen_reg, hdr_ready_seen_next;

reg                         m_eth_hdr_valid_int;
reg [47:0]                  m_eth_dest_mac_int;
reg [47:0]                  m_eth_src_mac_int;
reg [15:0]                  m_eth_type_int;
reg [DATA_WIDTH-1:0]        m_eth_payload_axis_tdata_int;
reg [KEEP_WIDTH-1:0]        m_eth_payload_axis_tkeep_int;
reg                         m_eth_payload_axis_tvalid_int;
reg                         m_eth_payload_axis_tlast_int;
reg                         m_eth_payload_axis_tuser_int;

reg [DATA_WIDTH-1:0]        m_data_fifo_tdata;
reg [KEEP_WIDTH-1:0]        m_data_fifo_tkeep;
reg                         m_data_fifo_tvalid;
reg                         m_data_fifo_tready;
reg                         m_data_fifo_tlast;
reg                         m_data_fifo_tuser;

assign m_eth_hdr_valid = m_eth_hdr_valid_int;
assign m_eth_dest_mac = m_eth_dest_mac_int;
assign m_eth_src_mac = m_eth_src_mac_int;
assign m_eth_type = m_eth_type_int;
assign m_eth_payload_axis_tdata = m_eth_payload_axis_tdata_int;
assign m_eth_payload_axis_tkeep = m_eth_payload_axis_tkeep_int;
assign m_eth_payload_axis_tvalid = m_eth_payload_axis_tvalid_int;
assign m_eth_payload_axis_tuser = m_eth_payload_axis_tuser_int;
assign m_eth_payload_axis_tlast = m_eth_payload_axis_tlast_int;

localparam [2:0]
    STATE_CHECK     = 3'd0,
    STATE_BYPASS    = 3'd1,
    STATE_STORE     = 3'd2,
    STATE_WAIT      = 3'd3,
    STATE_DUPLICATE = 3'd4;

reg [2:0] state_reg = STATE_CHECK, state_next;

// fifo doesn't have an enable, so we gate tvalid.
reg s_data_fifo_tvalid;
reg s_data_fifo_tready;

axis_fifo #(
    .DEPTH(2048/(DATA_WIDTH/8)),    // 2048 bytes,
    .DATA_WIDTH(DATA_WIDTH),
    .FRAME_FIFO(1)                  // store whole packet
) u_data_fifo (
    .clk                (clk),
    .rst                (rst),

    .s_axis_tdata       (s_eth_payload_axis_tdata),
    .s_axis_tkeep       (s_eth_payload_axis_tkeep),
    .s_axis_tvalid      (s_data_fifo_tvalid),
    .s_axis_tready      (s_data_fifo_tready),
    .s_axis_tlast       (s_eth_payload_axis_tlast),
    .s_axis_tid         (1'b0),
    .s_axis_tdest       (1'b0),
    .s_axis_tuser       (s_eth_payload_axis_tuser),

    .m_axis_tdata       (m_data_fifo_tdata),
    .m_axis_tkeep       (m_data_fifo_tkeep),
    .m_axis_tvalid      (m_data_fifo_tvalid),
    .m_axis_tready      (m_data_fifo_tready),
    .m_axis_tlast       (m_data_fifo_tlast),
    .m_axis_tid         (),
    .m_axis_tdest       (),
    .m_axis_tuser       (m_data_fifo_tuser),

    .pause_req          ('0),
    .pause_ack          (),

    .status_depth       (),
    .status_depth_commit(),
    .status_overflow    (),
    .status_bad_frame   (),
    .status_good_frame  ()
);

wire source_mac_match;
wire dest_mac_match;
wire ethertype_match;

filter #(
    .WIDTH(48)
) u_source_mac_filter (
    .filter     (source_mac_filter),
    .mask       (source_mac_filter_mask),
    .data       (s_eth_src_mac),
    .enable     (source_mac_filter_en),
    .match      (source_mac_match)
);

filter #(
    .WIDTH(48)
) u_dest_mac_filter (
    .filter     (dest_mac_filter),
    .mask       (dest_mac_filter_mask),
    .data       (s_eth_dest_mac),
    .enable     (dest_mac_filter_en),
    .match      (dest_mac_match)
);

filter #(
    .WIDTH(15)
) u_ethertype_filter (
    .filter     (ethertype_filter),
    .mask       (ethertype_filter_mask),
    .data       (s_eth_type),
    .enable     (ethertype_filter_en),
    .match      (ethertype_match)
);

modifier #(
    .WIDTH(48)
) u_source_mac_modifier (
    .modifier   (source_mac_mod),
    .mask       (source_mac_mod_mask),
    .data_in    (source_mac_reg),
    .enable     (source_mac_mod_en),
    .data_out   (source_mac_modified)
);

modifier #(
    .WIDTH(48)
) u_dest_mac_modifier (
    .modifier   (dest_mac_mod),
    .mask       (dest_mac_mod_mask),
    .data_in    (dest_mac_reg),
    .enable     (dest_mac_mod_en),
    .data_out   (dest_mac_modified)
);

modifier #(
    .WIDTH(15)
) u_ethertype_modifier (
    .modifier   (ethertype_mod),
    .mask       (ethertype_mod_mask),
    .data_in    (ethertype_reg),
    .enable     (ethertype_mod_en),
    .data_out   (ethertype_modified)
);

always @* begin
    state_next = state_reg;

    s_eth_hdr_ready_int = 1'b0;
    s_data_fifo_tvalid = 1'b0;
    m_eth_hdr_valid_int = 1'b0;

    m_eth_src_mac_int = 48'b0;
    m_eth_dest_mac_int = 48'b0;
    m_eth_type_int = 16'b0;

    m_eth_payload_axis_tdata_int = {DATA_WIDTH{1'b0}};
    m_eth_payload_axis_tkeep_int = 1'b0;
    m_eth_payload_axis_tvalid_int = 1'b0;
    m_eth_payload_axis_tuser_int = 1'b0;
    m_eth_payload_axis_tlast_int = 1'b0;

    s_eth_payload_axis_tready_int = 1'b0;

    m_data_fifo_tready = 1'b0;

    source_mac_next = source_mac_reg;
    dest_mac_next = dest_mac_reg;
    ethertype_next = ethertype_reg;

    hdr_ready_seen_next = '0;

    case (state_reg)
        STATE_CHECK: begin

            m_eth_hdr_valid_int = 1'b0;
            s_eth_hdr_ready_int = 1'b0;
            m_eth_src_mac_int = s_eth_src_mac;

            m_eth_payload_axis_tvalid_int = 1'b0;
            s_eth_payload_axis_tready_int = 1'b0;
            m_eth_payload_axis_tlast_int = 1'b0;

            if (s_eth_hdr_valid) begin
                if (source_mac_match | dest_mac_match | ethertype_match) begin
                    state_next = STATE_WAIT;
                end else begin
                    state_next = STATE_BYPASS;
                end
            end
        end
        STATE_BYPASS: begin
            m_eth_hdr_valid_int = s_eth_hdr_valid;
            s_eth_hdr_ready_int = m_eth_hdr_ready;
            m_eth_src_mac_int = s_eth_src_mac;
            m_eth_dest_mac_int = s_eth_dest_mac;
            m_eth_type_int = s_eth_type;
            m_eth_payload_axis_tdata_int = s_eth_payload_axis_tdata;
            m_eth_payload_axis_tkeep_int = s_eth_payload_axis_tkeep;
            m_eth_payload_axis_tvalid_int = s_eth_payload_axis_tvalid;
            s_eth_payload_axis_tready_int = m_eth_payload_axis_tready;
            m_eth_payload_axis_tuser_int = s_eth_payload_axis_tuser;
            m_eth_payload_axis_tlast_int = s_eth_payload_axis_tlast;

            if (m_eth_payload_axis_tvalid & m_eth_payload_axis_tready & m_eth_payload_axis_tlast) begin
                state_next = STATE_CHECK;
            end
        end
        STATE_WAIT: begin
            m_eth_hdr_valid_int = s_eth_hdr_valid;
            s_eth_hdr_ready_int = m_eth_hdr_ready;
            m_eth_src_mac_int = s_eth_src_mac;
            m_eth_dest_mac_int = s_eth_dest_mac;
            m_eth_type_int = s_eth_type;
            m_eth_payload_axis_tdata_int = s_eth_payload_axis_tdata;
            m_eth_payload_axis_tkeep_int = s_eth_payload_axis_tkeep;
            m_eth_payload_axis_tvalid_int = s_eth_payload_axis_tvalid;
            s_eth_payload_axis_tready_int = m_eth_payload_axis_tready;
            m_eth_payload_axis_tuser_int = s_eth_payload_axis_tuser;
            m_eth_payload_axis_tlast_int = s_eth_payload_axis_tlast;
            s_data_fifo_tvalid = s_eth_payload_axis_tvalid & s_eth_payload_axis_tready;

            if (s_eth_hdr_valid) begin
                source_mac_next = s_eth_src_mac;
                dest_mac_next = s_eth_dest_mac;
                ethertype_next = s_eth_type;
            end


            if (m_eth_payload_axis_tvalid & m_eth_payload_axis_tready & m_eth_payload_axis_tlast) begin
                state_next = STATE_DUPLICATE;
            end
        end
        STATE_DUPLICATE: begin
            hdr_ready_seen_next = hdr_ready_seen_reg;
            if (!hdr_ready_seen_reg) begin
                m_eth_hdr_valid_int = 1'b1;
                hdr_ready_seen_next = m_eth_hdr_ready;
                m_eth_dest_mac_int = dest_mac_modified;
                m_eth_src_mac_int = source_mac_modified;
                m_eth_type_int = ethertype_modified;
            end

            m_eth_payload_axis_tdata_int = m_data_fifo_tdata;
            m_eth_payload_axis_tkeep_int = m_data_fifo_tkeep;
            m_eth_payload_axis_tvalid_int = m_data_fifo_tvalid;
            m_data_fifo_tready = m_eth_payload_axis_tready;
            m_eth_payload_axis_tlast_int = m_data_fifo_tlast;
            m_eth_payload_axis_tuser_int = m_data_fifo_tuser;

            if (m_eth_payload_axis_tvalid & m_eth_payload_axis_tready & m_eth_payload_axis_tlast) begin
                state_next = STATE_CHECK;
            end
        end
    endcase
end

always @(posedge clk) begin
    if (rst) begin
        state_reg <= STATE_CHECK;
        source_mac_reg <= 48'b0;
        dest_mac_reg <= 48'b0;
        ethertype_reg <= 16'b0;
        hdr_ready_seen_reg <= 1'b0;
    end else begin
        state_reg <= state_next;

        source_mac_reg <= source_mac_next;
        dest_mac_reg <= dest_mac_next;
        ethertype_reg <= ethertype_next;

        hdr_ready_seen_reg <= hdr_ready_seen_next;
    end
end


endmodule